package LEARNAgents;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.apache.logging.log4j.ThreadContext;

import PlatformSelector.MyPlatformDescription;
import jade.content.lang.Codec.CodecException;
import jade.content.onto.OntologyException;
import jade.core.AID;
import jade.core.Agent;
import jade.core.behaviours.CyclicBehaviour;
import jade.core.behaviours.OneShotBehaviour;
import jade.lang.acl.ACLMessage;
import jade.lang.acl.MessageTemplate;

/**
   This example shows an agent able to respond to other agents wishing
   to know if it is alive. More in details we use a 
   <code>CyclicBehaviour</code> that receives only messages of type 
   <code>QUERY_IF</code> and using the <code>presence</code> ontology.
   All other messages are ignored. This is achieved by specifying a 
   proper <code>MessageTemplate</code>. Whenever such a message is 
   received an <code>INFORM</code> message is sent back as reply.
   @author Giovanni Caire - TILAB
 */
public class RequesterAgent extends Agent {
	//private MessageTemplate template = MessageTemplate.and(
	//	MessageTemplate.MatchPerformative(ACLMessage.QUERY_IF),
	//	MessageTemplate.MatchOntology("presence") );
		private Logger logger;// = LogManager.getLogger(LEARNAMS.class);
        private int numServ=3;
        private int received=0;
        
	protected void setup() {
		
		ThreadContext.put("ROUTINGKEY", this.getName());
		logger = LogManager.getLogger(this.getName());
		
		Object[] args = getArguments();
	  	if (args != null && args.length > 0) {
	  		numServ = (Integer.parseInt((String)args[0]));
	  	}
		  
		addBehaviour(new CyclicBehaviour(this) {
			public void action() {
				try{
					ACLMessage msg = myAgent.receive();
					if (msg != null) {
						System.out.println("Requester:" + getName() + " Received message from agent "+msg.getSender().getName() + " content: " + msg.getContent());
						logger.trace("Message from " + msg.getSender().getName() + " content: " + msg.getContent());
						received++;
						
						if(received==numServ){
							System.out.println("Requester:" + getName() + " RECEIVED ALL ANSWERS AT " + System.currentTimeMillis());
							logger.trace("Requester " + getName() + ": RECEIVED ALL ANSWERS AT " + System.currentTimeMillis());
							//doDelete();
						}
					}
					else {
						block();
					}
				}
				catch(Exception e){
					logger.trace("I would have die because " + e.getMessage());
					System.out.println(myAgent.getName() + " sarei morto perche' " + e.getMessage());
				}
				
			}
		} );
		
		addBehaviour(new OneShotBehaviour(this) {
			public void action() {
				ACLMessage mess=new ACLMessage(ACLMessage.REQUEST);
				mess.addReceiver(new AID("SM",false)); //(new AID("SM@Platform1",true));
				mess.setProtocol("Invoke");
				//mess.setLanguage(codec.getName());
				//mess.setOntology(ontology.getName());
				mess.setContent("1;s1;param;"+ numServ +";true");
				logger.trace("ASK GROUPINVOKE WITH PARAM: 1;s1;param;"+ numServ +";true at " + System.currentTimeMillis());
				System.out.println("Requester: ASK GROUPINVOKE WITH PARAM: 1;s1;param;" + numServ + ";true at " + System.currentTimeMillis());
				send(mess);
			}
		} );
	}
}
