package LEARNAgents;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.apache.logging.log4j.ThreadContext;

import LEARNOntology.Execute;
import LEARNOntology.GetKnownPlatforms;
import LEARNOntology.LEARNOntology;
import LEARNOntology.LEARNServiceDescription;
import LEARNOntology.WorkFor;
import LEARNOntology.impl.DefaultPlatformDescription;
import LEARNOntology.impl.DefaultS1;
import LEARNOntology.impl.DefaultWorkFor;
import PlatformSelector.MyPlatformDescription;
import jade.content.Concept;
import jade.content.ContentElement;
import jade.content.lang.Codec;
import jade.content.lang.Codec.CodecException;
import jade.content.lang.sl.SLCodec;
import jade.content.onto.Ontology;
import jade.content.onto.OntologyException;
import jade.content.onto.basic.Action;
import jade.core.AID;
import jade.core.Agent;
import jade.core.behaviours.CyclicBehaviour;
import jade.domain.DFService;
import jade.domain.FIPAException;
import jade.domain.FIPAAgentManagement.DFAgentDescription;
import jade.domain.FIPAAgentManagement.SearchConstraints;
import jade.domain.FIPAAgentManagement.ServiceDescription;
import jade.lang.acl.ACLMessage;
import jade.lang.acl.MessageTemplate;
import jade.util.leap.Iterator;
import jade.wrapper.AgentController;
import jade.wrapper.ContainerController;
import jade.wrapper.StaleProxyException;

public class ServiceManager extends Agent{
	
	private Codec codec = new SLCodec(); //FIPANames.ContentLanguage.FIPA_SL0; 
	private Ontology ontology = LEARNOntology.getInstance();
	
	private MessageTemplate templateExecute = MessageTemplate.MatchProtocol("execute");
	private MessageTemplate templateInvoke = MessageTemplate.MatchProtocol("invoke");
	
	private Logger logger;// = LogManager.getLogger(LEARNAMS.class);
    
	protected void setup() {
		
		ThreadContext.put("ROUTINGKEY", this.getName());
		logger = LogManager.getLogger(this.getName());
		  
		getContentManager().registerLanguage(codec); //, FIPANames.ContentLanguage.FIPA_SL0);	
		// Register the ontology used by this application
		getContentManager().registerOntology(ontology);
		
		addBehaviour(new CyclicBehaviour(this) {
			public void action() {
				ACLMessage msg = myAgent.receive(templateExecute);
				if (msg != null) {
					if(!msg.getSender().getName().startsWith("ams")){
						// mi stanno chiedendo di iniziare un protocollo di execute
						if(msg.getPerformative()==ACLMessage.REQUEST){
							System.out.println("I'm " + myAgent.getName() + ". Received execute from agent "+msg.getSender().getName());
							logger.trace("Received execute from agent "+msg.getSender().getName());
							execute(msg);
						}
						/*
						else if(msg.getPerformative()==ACLMessage.INFORM){
							
						}*/
						
					}
				}
				
				msg = myAgent.receive(templateInvoke);
				if (msg != null) {
					if(msg.getSender().getName().startsWith("Assistant") && msg.getPerformative()==ACLMessage.INFORM){
						// questo è uno degli assistant che mi dice che ha finito
						// per ora ne prendo solo atto
						System.out.println("I'm SM and received completed protocol from " +msg.getSender().getName());
						logger.trace("Received completed protocol from " +msg.getSender().getName());
					}
					else if(!msg.getSender().getName().startsWith("ams")){
						
						// workFor(LEARNNodeA,ID,ServiceName,[ServiceParams],NumberOfAnswers,Singularly)
	                	// example
						// 1;s1;weqew;3;true
						
	                	String arguments=msg.getSender().getLocalName() + ";" + msg.getContent();
	                	// se non uso l'ID nel nome dell'assistant, mi rimane un assistant per nodo invece che per richiesta,
	                	// ed e' un problema perche' poi i LN non mi rispondono piu' per diverse richieste da stesso nodo (che avrebbe lo stesso Assistant)
	                	String[] params=msg.getContent().split(";");
	                	String id=params[0];
	                	
	                	LEARNServiceDescription sd= createServiceDescription(params);

	                	DefaultWorkFor wf=new DefaultWorkFor();
	                	wf.setRequestID(Integer.parseInt(id));
	                	wf.setLearnNode(msg.getSender());
	                	wf.setSingularly(Boolean.parseBoolean(params[4]));
	                	wf.setRequestedService(sd);
	                	wf.setNumRequestedResults(Integer.parseInt(params[3]));
	                	
						ContainerController cc = getContainerController();
						AgentController ac;
						try {
							ac = cc.createNewAgent("Assistant_"+ msg.getSender().getLocalName() + "_" + id, "LEARNAgents.Assistant", null);
							ac.start();
							
						} catch (StaleProxyException e) {
							// TODO Auto-generated catch block
							e.printStackTrace();
						}

						ACLMessage workF=new ACLMessage(ACLMessage.INFORM);
						//workF.setContent(arguments);
						workF.addReceiver(new AID("Assistant_"+ msg.getSender().getLocalName() + "_" + id, false));
						
						workF.setLanguage(codec.getName());
						workF.setOntology(ontology.getName());
										
						try {
							Action a = new Action();
							
							// DANY  altrimenti temo che i learnnode non siano poi in grodi di rispondermi, se su più piattaforme
							//a.setActor(new AID(getLocalName(), AID.ISLOCALNAME));
							AID workForGlobal= new AID(getName(),true);
							try{
								workForGlobal.addAddresses(myAgent.getAID().getAddressesArray()[0]);	
									
								}
							catch(Exception e){
								logger.trace("exception, no address found in me");
							}
							
							try{
									
								workForGlobal.addAddresses(myAgent.getAMS().getAllAddresses().next().toString());	
								}
							catch(Exception e){
								logger.trace("exception, no address found in AMS");
							}
							
							a.setActor(workForGlobal);
							//
							
							a.setAction(wf);
							getContentManager().fillContent(workF,a);				
							send(workF);		
							
						} catch (CodecException | OntologyException e) {
							// TODO Auto-generated catch block
							e.printStackTrace();
							logger.trace("errore in sending mess " + workF.toString());
							logger.trace(e.getMessage());
							
							System.out.println("errore in sending mess " + workF.toString());
						} 
						
						logger.trace("Assistant created:" + "Assistant_"+ msg.getSender().getLocalName() + "_" + id);
						// qui potrei informare il mio LEARNAMS di una nuova piattaforma (vede lui se e' nuova)
					}
					/* else{
						
					}*/
				}
			}

			private LEARNServiceDescription createServiceDescription(String[] params) {
				LEARNServiceDescription serv=null;
				switch(params[1]) {
					case "s1":
						serv=new DefaultS1();
						serv.setServiceName(params[1]);
						break;
					
					default:
						break;
				}
				
				// qui non so bene come farlo, ogni servizio avra' una sua lista di parametri, di un certo tipo...
				// ma questo pezzo di codice in realtà non va qui nel SM, ma nel LN, che manderà secco la richiesta con dentro un servizio...
				// quindi qui per ora creo un servizio solo con il nome, senza parametri... poi ci porremo il problem
				
				return serv;
				
			}
		} );
		
	}
	
	private void execute(ACLMessage msg){
		ContentElement ce = null; 
		try {
			ce=getContentManager().extractContent(msg);
			Concept c = (Concept)((Action)ce).getAction();
			
			// deve essere questo, ma per sicurezza...
			 if(c instanceof Execute){
				LEARNServiceDescription serv=((Execute)c).getRequestedService();
				int numb=((Execute)c).getNumRequestedResults();
				
				DFAgentDescription template = new DFAgentDescription();
			  	ServiceDescription templateSd = new ServiceDescription();
			  	templateSd.setName(serv.getServiceName());
			  	templateSd.setType("LEARN");
			  	// per ora lasciamo un attimo stare i parametri del servizio
			  	template.addServices(templateSd);
			  		
			  	SearchConstraints sc = new SearchConstraints();
			  	
			  	sc.setMaxResults(new Long(numb));
			  	DFAgentDescription[] results;
				try {
					System.out.println("cerco " + numb  + " servizi di nome " + serv.getServiceName());
					results = DFService.search(this, template, sc);
					System.out.println("ne ho trovati " + results.length);
					logger.trace("Found " + results.length + " of type " + serv.getServiceName() + "(requested were:" + numb);
					for (int i = 0; i < results.length; ++i) {
						DFAgentDescription dfd = results[i];
						AID provider = dfd.getName();
						// The same agent may provide several services
						Iterator it = dfd.getAllServices();
						while (it.hasNext() ) {
							ServiceDescription sd = (ServiceDescription) it.next();
							if (sd.getName().equals(serv.getServiceName()) && sd.getType().equals("LEARN")){//!sdlist.contains(sd)) {
								logger.trace("Found Service (by local DF) \""+sd.getName()+"\" provided by agent "+provider.getName() + ": I invoke it for Assistant "); // + msg.getSender().getAddressesArray()[0]);
								
								invoke(provider,serv,msg.getSender());
								
							}
						}
					}
					
				} catch (FIPAException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}
			 }
		} catch (CodecException | OntologyException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}
	
	//vecchia
	private void execute(String param){
		
		// mandare un mess al LEARNAMS se la piattaforma e l'ip non sono quelli locali
		// TBD
		
		// qui la sintassi è:
		// execute(ServiceName;[ServiceParams con ,];(NumberOfAnswer-#messages)/#PlaformsList;AssistantA@Platform1;IP)
		// ora mi arriva una azione di tipo execute
		
		String[] params=param.split(";");
		
		DFAgentDescription template = new DFAgentDescription();
  		ServiceDescription templateSd = new ServiceDescription();
  		templateSd.setName(params[0]);
  		templateSd.setType("LEARN");
  		// per ora lasciamo un attimo stare i parametri del servizio
  		template.addServices(templateSd);
  		
  		SearchConstraints sc = new SearchConstraints();
  		
  		sc.setMaxResults(new Long(params[2]));
  		
  		DFAgentDescription[] results;
		try {
			System.out.println("cerco i servizi di nome " + params[0]);
			results = DFService.search(this, template, sc);
			System.out.println("ne ho trovati " + results.length);
			logger.trace("Found " + results.length + " of type " + params[0]);
			for (int i = 0; i < results.length; ++i) {
				DFAgentDescription dfd = results[i];
				AID provider = dfd.getName();

				Iterator it = dfd.getAllServices();
				while (it.hasNext() ) {
					ServiceDescription sd = (ServiceDescription) it.next();
					if (sd.getName().equals(params[0]) && sd.getType().equals("LEARN")){//!sdlist.contains(sd)) {
						logger.trace("Found Service (by local DF) \""+sd.getName()+"\" provided by agent "+provider.getName());
						
						invoke(provider,sd,params[1],params[3],params[4]);
						
					}
				}
			}
			
		} catch (FIPAException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
  		
	}
	
	// vecchia
	private void invoke(AID provider,ServiceDescription sd, String params, String srequester, String IP){
		// qui dentro ci andrà un grosso if, in base al servizio richiesto, che sappiamo come invocare
		if(sd.getName().equals("s1")){
			// o il performative corretto, dipende dal servizio
			ACLMessage mess=new ACLMessage(ACLMessage.REQUEST);
			mess.addReceiver(provider);
			
			AID requester= new AID(srequester,true);
			requester.setName(srequester);
			requester.addAddresses(IP);
			//requester.ISGUID(true);
			
			mess.addReplyTo(requester);
			
			// qui dipende da come imposta la richiesta di servizio, per ora facciamolo stupido
			mess.setContent(sd.getName());
			//mess.addUserDefinedParameter(key, value);
			send(mess);
		}
		
	}
	
	private void invoke(AID provider,LEARNServiceDescription sd, AID respondTo){
		// qui dentro ci andrà un grosso if, in base al servizio richiesto, che sappiamo come invocare
		
		// o il performative corretto, dipende dal servizio
		ACLMessage mess=new ACLMessage(ACLMessage.REQUEST);
		mess.addReceiver(provider);
		mess.addReplyTo(respondTo);
		
		if(sd.getServiceName().equals("s1")){

			// qui dipende da come imposta la richiesta di servizio, per ora facciamolo stupido
			mess.setContent(sd.getServiceName() + ";");
			//mess.addUserDefinedParameter(key, value);
			send(mess);
		}
		
	}


}
